/*
 *  Project   : SPECTRUM - Digital Spectrometric System of Nuclear Radiation
 *
 *  Tool      : ClassificatorScript - spectrometric data processing
 *               
 *  Author(s) : Martin VESKRNA <256569@mail.muni.cz>             
 *
 *  Date      : 
 *
 *  (c) Copyright 2011 MASARYK UNIVERSITY
 *  All rights reserved.
 *
 *  Please review the terms of the license agreement before using this
 *  file. If you are not an authorized user, please destroy this
 *  source code file and notify MASARYK UNIVERSITY immediately that you
 *  inadvertently received an unauthorized copy.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include "udp.h"
#include "constants.h"





/********************************************************
/     Output UDP
********************************************************/

unsigned int UDPoutputPacketNumber;
struct sockaddr_in si_otherOutput;
int socketOutput, slenOutput=sizeof(si_otherOutput);

int PrepareOutputUDP(char* IP, int port)
{
	if ((socketOutput=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP))==-1)
	{
		fprintf(stderr, "Can not create socket for UDP output transferee\n");
		return 0;
	}

	memset((char *) &si_otherOutput, 0, sizeof(si_otherOutput));
	si_otherOutput.sin_family = AF_INET;
	si_otherOutput.sin_port = htons(port);
	
	if (inet_aton(IP, &si_otherOutput.sin_addr)==0)
	{
		fprintf(stderr, "inet_aton() failed - Spatny format IP adresy\n");
		return 0;
	}
	
	UDPoutputPacketNumber = 0;
	
	return 1;
}

void CloseOutputUDP()
{
	close(socketOutput);
}

#define PACKET_LENGTH 22
char buf[PACKET_LENGTH];

int SendUDPpacket(char* buffer, unsigned int length)
{
	if (sendto(socketOutput, buffer, length, 0, (struct sockaddr*)&si_otherOutput, slenOutput) == -1)
    {
		fprintf(stderr, "Error while UDP packet sending\n");
		return 0;
	}
	
	return 1;

}



int SendUDPmethodPacket(unsigned long long time, unsigned int pulseNum, unsigned short valueA, unsigned short valueB)
{
	unsigned short pulseCount = 1;

	//packet construction
	memcpy(buf, (void *) &UDPoutputPacketNumber, 4);		//Packet number
	memcpy(buf+4, (void *) &pulseCount, 2);					//pulse count
	memcpy(buf+6, (void *) &pulseNum, 4);			//ADC packet number
	memcpy(buf+10, (void *) &time, 8);				//ADC packet time
	memcpy(buf+18, (void *) &valueA, 2);			//firts parameter - amplitude
	memcpy(buf+20, (void *) &valueB, 2);			//second parameter - discrimination parameter
	
	if (SendUDPpacket(buf, PACKET_LENGTH) == 0)
	{
		return 0;
	}
	
	UDPoutputPacketNumber++;
	return 1;
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////






/********************************************************
/     Input UDP
********************************************************/

//computed by spectromeasure
double hwPeriod_ns = 7.957855135;	//TODO - take from global .h file


int socketInput;
struct sockaddr_in servaddr, cliaddr;
socklen_t slenInput = sizeof(cliaddr);
#define INPUT_BUF_LEN 2000
unsigned char inputBuf[INPUT_BUF_LEN];
#define TIMEOUT_SEC 1

int PrepareInputUDP(int port)
{
	if ((socketInput=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
    {
		fprintf(stderr, "Can not create socket for UDP input transferee\n");
		return 0;
    }
	
	//set timer for recv_socket
	struct timeval tv;
	tv.tv_sec = TIMEOUT_SEC;
	tv.tv_usec = 0;
	setsockopt(socketInput, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv));
   
    memset((char *) &servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(port);
   
    if (bind(socketInput, (struct sockaddr *)&servaddr, sizeof(servaddr)) == -1)
	{
		fprintf(stderr, "Can not bind input port\n");
		return 0;
	}
	

	
	return 1;
}

void CloseInputUDP()
{
	close(socketInput);
}

unsigned short GetSampleA(unsigned char *data, int i)
{
	return ((unsigned short)data[i] << 4 ) | ((unsigned short)data[i+1] >> 4 );
}

unsigned short GetSampleB(unsigned char *data, int i)
{
	return (((unsigned short)data[i+1] & 0xF) << 8 ) | (unsigned short)data[i+2];
}


unsigned int prevPacketNum = 0;
unsigned int prevPulseNum = 0;
unsigned long long prevTime = 0;
int prevDelka = 0;
int firstRun = 1;

///Prijme po UDP obe vetve ADC prevodniku
///@CH1 - data vetve 1 (kanal 1)
///&CH2 - data vetve 2 (kanal 2)
///@delka - pocet vzorku pulzu
///@casImpulsu_ns - cas v ns kdy byl impulz ADC kartou zachycen
///@pulzKompletni - nechybi zadna cast pulzu (neni implementovano, nutno dodat do .dat souboru)
///@vypadek_ns - doba vypadku v ns pred pulzem
///@vypadek_pulzy - pocet zpracenych pulzu
int ReceiveRAWdataPacket(float* CH1, float* CH2, int* delka, unsigned long long* casImpulsu_ns, int* pulzKompletni, unsigned long long* vypadek_ns, int* vypadek_pulzy)
{
	int n;
	n = recvfrom(socketInput, inputBuf, INPUT_BUF_LEN, 0, (struct sockaddr *)&cliaddr, &slenInput);
	
	if (n == -1 && errno == 11)
	{
		//no message income - timeout
		return 0;
	}
	else if (n == -1)
	{
		fprintf(stderr, "Error while UDP packet receiving\n");
		return 0;
	}
	
	
	//length of ethernet header
	#define PACKET_HEADER_ETHERNET_LENGTH (6+6+2)
	//length of oscilo packet header
	#define PACKET_HEADER_LENGTH (PACKET_HEADER_ETHERNET_LENGTH+1+4+8+4+1)
	//length of oscilo packet footer
	#define PACKET_FOOTER_LENGTH (1)
	
	#define HW_SAMPLES_IN_1_PERIOD 16
	
	//check packet length
	if ((n - PACKET_HEADER_LENGTH - PACKET_FOOTER_LENGTH) % 12*HW_SAMPLES_IN_1_PERIOD/8 != 0)
	{
		fprintf(stderr, "error - packet length is not correct. Lenght is %d\n", *delka);
		return 0;
	}
	
	unsigned int packetNum =	((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+1] << 24 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+2] << 16 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+3] << 8 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+4]);

	unsigned long long HW_timeStamp = ((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+5] << 56 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+6] << 48 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+7] << 40 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+8] << 32 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+9] << 24 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+10] << 16 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+11] << 8 ) |
			((unsigned long long)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+12]);

	unsigned int pulseNum = ((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+13] << 24 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+14] << 16 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+15] << 8 ) |
			((unsigned int)inputBuf[PACKET_HEADER_ETHERNET_LENGTH+16]);

	unsigned char FP_LP = inputBuf[n - 1];
	
	
	int i = PACKET_HEADER_LENGTH;
	*delka = 0;
	while (i < n - PACKET_FOOTER_LENGTH && *delka < MAX_IMPULSE_SIZE)
	{
		unsigned short sampleA = GetSampleA(inputBuf, i);
		unsigned short sampleB = GetSampleB(inputBuf, i);

		CH2[*delka] = (float)sampleA;
		CH1[*delka] = (float)sampleB;
		
		(*delka)++;
		i += 3;
	}
	
	*casImpulsu_ns = (unsigned long long)((double)HW_timeStamp * hwPeriod_ns);
	*pulzKompletni = (FP_LP == 0xFF) ? 1: 0;
	*vypadek_ns = 0;
	*vypadek_pulzy = 0;
	
	//loss detection
	if (packetNum != prevPacketNum + 1 && !firstRun)
	{
		*vypadek_pulzy = pulseNum - prevPulseNum - 1;
		if (*vypadek_pulzy == 0) *vypadek_pulzy = 1;
		
		*vypadek_ns = *casImpulsu_ns - (prevTime + prevDelka);
		//fprintf(stderr, "%llu, %d, %llu, %llu\n", prevTime, prevDelka, *casImpulsu_ns, *vypadek_ns);
	}
	
	
	//store previous value
	prevPacketNum = packetNum;
	prevPulseNum = pulseNum;
	prevTime = *casImpulsu_ns;
	prevDelka = *delka;
	firstRun = 0;
	
	return 1;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
